/*
Copyright (C) 2003-2004 Douglas Thain and the University of Wisconsin
Copyright (C) 2005- The University of Notre Dame
This software is distributed under a BSD-style license.
See the file COPYING for details.
*/

#include "chirp_acl.h"
#include "chirp_protocol.h"

#include "username.h"
#include "stringtools.h"
#include "debug.h"
#include "delete_dir.h"

#include <errno.h>
#include <string.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>

static int read_only_mode = 0;
static const char *default_acl=0;

static int do_stat( const char *filename, struct stat *buf )
{
	int result;
	do {
		result = stat(filename,buf);
	} while( result==-1 && errno==EINTR );
	return result;
}

void chirp_acl_force_readonly()
{
	read_only_mode = 1;
}

void chirp_acl_default( const char *d )
{
	default_acl = d;
}

static int is_a_directory( const char *filename )
{
	struct stat info;

	if(do_stat(filename,&info)==0) {
		if(S_ISDIR(info.st_mode)) {
			return 1;
		} else {
			return 0;
		}
	} else {
		return 0;
	}
}

static void make_acl_name( const char *filename, int get_parent, char *aclname )
{
	sprintf(aclname,"%s/%s",filename,CHIRP_ACL_BASE_NAME);
}

static int do_chirp_acl_get( const char *filename, const char *dirname, const char *subject )
{
	FILE *aclfile;
	char aclsubject[CHIRP_LINE_MAX];
	int aclflags, flags = 0;
  
	aclfile = chirp_acl_open(dirname);
  
	if(aclfile) {
		while(chirp_acl_read(aclfile,aclsubject,&aclflags)) {
			if(string_match(aclsubject,subject) ||
			   !strcmp(aclsubject,"system:anyuser")) {
				flags |= aclflags;
	
			}
		}
		chirp_acl_close(aclfile);
    
		if( read_only_mode ) {
			flags &= CHIRP_ACL_READ | CHIRP_ACL_LIST;
		}

		return flags;
    
	} else if(errno==ENOENT) {
		struct stat buf;
		if(do_stat(filename,&buf)==0) {
			if( buf.st_mode&0004 )
				flags |= CHIRP_ACL_READ | CHIRP_ACL_LIST;

			if( read_only_mode )
				return flags;

			if( buf.st_mode&0002 )
				flags |= CHIRP_ACL_WRITE;
			if( buf.st_mode&0001 )
				flags |= CHIRP_ACL_EXECUTE;
			return flags;
		} else {
			return 0;
		}
	}
	else {
		debug(D_NOTICE,"%s %s : %s\n",filename,dirname,strerror(errno));
		return 0;
	}
}

static int chirp_acl_get( const char *filename, const char *subject)
{
	char linkname[CHIRP_PATH_MAX];
	char temp[CHIRP_PATH_MAX];
	char dirname[CHIRP_PATH_MAX];
	int flags;

	int length = readlink(filename,linkname,sizeof(linkname));
	if(length>0) {
		linkname[length] = 0;
		debug(D_DEBUG,"symlink %s points to %s",filename,linkname);
		filename = linkname;
	}

	sprintf(temp,"%s/..",filename);
	string_collapse_path(temp,dirname,1);

	flags = do_chirp_acl_get(filename,dirname,subject);

	if(!strcmp(string_back(filename,CHIRP_ACL_BASE_LENGTH),CHIRP_ACL_BASE_NAME)) {
		if(flags&( CHIRP_ACL_WRITE | CHIRP_ACL_DELETE ) && !( flags & CHIRP_ACL_ADMIN ) ) {
			flags &= ~(CHIRP_ACL_WRITE | CHIRP_ACL_DELETE);
		}
	}

	return flags;

}

int chirp_acl_check( const char *filename, const char *subject, int flags )
{
	char linkname[CHIRP_PATH_MAX];
	char temp[CHIRP_PATH_MAX];
	char dirname[CHIRP_PATH_MAX];
	int myflags;

	int length = readlink(filename,linkname,sizeof(linkname));
	if(length>0) {
		linkname[length] = 0;
		debug(D_DEBUG,"symlink %s points to %s",filename,linkname);
		filename = linkname;
	}

	if(!strcmp(string_back(filename,CHIRP_ACL_BASE_LENGTH),CHIRP_ACL_BASE_NAME)) {
		if(flags&(CHIRP_ACL_WRITE | CHIRP_ACL_DELETE) ) {
			flags &= ~(CHIRP_ACL_WRITE | CHIRP_ACL_DELETE);
			flags |= CHIRP_ACL_ADMIN;
		}
	}

	sprintf(temp,"%s/..",filename);
	string_collapse_path(temp,dirname,1);

	myflags =  do_chirp_acl_get( filename, dirname, subject );

	if( ( myflags & flags ) == flags )
		return 1;
	else {
		errno = EACCES;
		return 0;
	}
}

int chirp_acl_check_dir( const char *dirname, const char *subject, int flags )
{
	int myflags = do_chirp_acl_get(dirname,dirname,subject);

	if( ( flags & myflags ) == flags )
		return 1;
	else {
		errno = EACCES;
		return 0;
	}
}

int chirp_acl_set( const char *dirname, const char *subject, int flags )
{
	char aclname[CHIRP_PATH_MAX];
	char newaclname[CHIRP_PATH_MAX];
	char aclsubject[CHIRP_LINE_MAX];
	int aclflags;
	FILE *aclfile, *newaclfile;
	int result;
	int replaced_acl_entry=0;

	if(!is_a_directory(dirname)) {
		errno = ENOTDIR;
		return -1;
	}

	sprintf(aclname,"%s/%s",dirname,CHIRP_ACL_BASE_NAME);
	sprintf(newaclname,"%s/%s.%d",dirname,CHIRP_ACL_BASE_NAME,getpid());

	aclfile = fopen(aclname,"r");

	/* If the acl never existed, then we can simply create it. */
	if(!aclfile && errno==ENOENT) {
		if(default_acl) {
			aclfile = fopen(default_acl,"r");
		} else {
			aclfile = fopen("/dev/zero","r");
		}
	}

	if(!aclfile) {
		errno = EACCES;
		return -1;
	}

	replaced_acl_entry = 0;

	newaclfile = fopen(newaclname,"w");
	if(!newaclfile) {
		fclose(aclfile);
		errno = EACCES;
		return -1;
	}

	while(chirp_acl_read(aclfile,aclsubject,&aclflags)) {
		if(!strcmp(subject,aclsubject)) {
			aclflags = flags;
			replaced_acl_entry = 1;
		}
		if(aclflags!=0) {
			fprintf(newaclfile,"%s %s\n",aclsubject,chirp_acl_flags_to_text(aclflags));
		}
	}

	if(!replaced_acl_entry) {
		fprintf(newaclfile,"%s %s\n",subject,chirp_acl_flags_to_text(flags));
	}

	if(ferror(newaclfile)) {
		errno = EACCES;
		result = -1;
	} else {
		result = rename(newaclname,aclname);
		if(result<0) {
			unlink(newaclname);
			errno = EACCES;
			result = -1;
		}
	}

	fclose(aclfile);
	fclose(newaclfile);

	return result;
}


FILE * chirp_acl_open( const char *dirname )
{
	char aclname[CHIRP_PATH_MAX];
	FILE *file;

	if(!is_a_directory(dirname)) {
		errno = ENOENT;
		return 0;
	} else {
		make_acl_name(dirname,0,aclname);
		file = fopen(aclname,"r");
		if(!file && default_acl) file = fopen(default_acl,"r");
		return file;
	}
}

int chirp_acl_read( FILE *aclfile, char *subject, int *flags )
{
	char acl[CHIRP_LINE_MAX];
	char tmp[CHIRP_LINE_MAX];

	while(fgets(acl,sizeof(acl),aclfile)) {
		if(sscanf(acl,"%[^ ] %[rwldvax()]",subject,tmp)==2) {
			*flags = chirp_acl_text_to_flags(tmp);
			return 1;
		} else {
			continue;
		}
	}

	return 0;
}

void chirp_acl_close( FILE *aclfile )
{
	fclose(aclfile);
}


const char * chirp_acl_flags_to_text( int flags )
{
	static char text[20];

	text[0] = 0;
	if(flags&CHIRP_ACL_READ)    strcat(text,"r");
	if(flags&CHIRP_ACL_WRITE)   strcat(text,"w");
	if(flags&CHIRP_ACL_LIST)    strcat(text,"l");
	if(flags&CHIRP_ACL_DELETE)  strcat(text,"d");
	if(flags&CHIRP_ACL_ADMIN)   strcat(text,"a");
	if(flags&CHIRP_ACL_EXECUTE) strcat(text,"x");
	if(flags&CHIRP_ACL_RESERVE) {
		strcat(text,"v");
		strcat(text,"(");
		if(flags&CHIRP_ACL_RESERVE_READ)    strcat(text,"r");
		if(flags&CHIRP_ACL_RESERVE_WRITE)   strcat(text,"w");
		if(flags&CHIRP_ACL_RESERVE_LIST)    strcat(text,"l");
		if(flags&CHIRP_ACL_RESERVE_DELETE)  strcat(text,"d");
		if(flags&CHIRP_ACL_RESERVE_RESERVE) strcat(text,"v");
		if(flags&CHIRP_ACL_RESERVE_ADMIN)   strcat(text,"a");
		if(flags&CHIRP_ACL_RESERVE_EXECUTE) strcat(text,"x");
		strcat(text,")");
	}

	return text;
}

int chirp_acl_text_to_flags( const char *t )
{
	int flags=0;

	while(*t) {
		if( *t=='r' ) flags|=CHIRP_ACL_READ;
		if( *t=='w' ) flags|=CHIRP_ACL_WRITE;
		if( *t=='l' ) flags|=CHIRP_ACL_LIST;
		if( *t=='d' ) flags|=CHIRP_ACL_DELETE;
		if( *t=='a' ) flags|=CHIRP_ACL_ADMIN;
		if( *t=='x' ) flags|=CHIRP_ACL_EXECUTE;
		if( *t=='v' ) {
			flags |= CHIRP_ACL_RESERVE;
			if( t[1] == '(' ) {
				t += 2;
				while(*t && *t != ')'){
					if( *t=='r' ) flags|=CHIRP_ACL_RESERVE_READ;
					if( *t=='w' ) flags|=CHIRP_ACL_RESERVE_WRITE;
					if( *t=='l' ) flags|=CHIRP_ACL_RESERVE_LIST;
					if( *t=='d' ) flags|=CHIRP_ACL_RESERVE_DELETE;
					if( *t=='v' ) flags|=CHIRP_ACL_RESERVE_RESERVE;
					if( *t=='a' ) flags|=CHIRP_ACL_RESERVE_ADMIN;
					if( *t=='x' ) flags|=CHIRP_ACL_RESERVE_EXECUTE;
					++t;
				}
			}
		}
		++t;
	}

	return flags;
}

int chirp_acl_from_access_flags( int flags )
{
	int acl=0;
	if(flags&R_OK) acl|=CHIRP_ACL_READ;
	if(flags&W_OK) acl|=CHIRP_ACL_WRITE;
	if(flags&X_OK) acl|=CHIRP_ACL_EXECUTE;
	if(flags&F_OK) acl|=CHIRP_ACL_READ;
	if(acl==0)     acl|=CHIRP_ACL_READ;
	return acl;
}

int chirp_acl_from_open_flags( int flags )
{
	int acl = 0;
	if(flags&O_WRONLY) acl|=CHIRP_ACL_WRITE;
	if(flags&O_RDWR)   acl|=CHIRP_ACL_READ|CHIRP_ACL_WRITE;
	if(flags&O_CREAT)  acl|=CHIRP_ACL_WRITE;
	if(flags&O_TRUNC)  acl|=CHIRP_ACL_WRITE;
	if(flags&O_APPEND) acl|=CHIRP_ACL_WRITE;
	if(acl==0)         acl|=CHIRP_ACL_READ;
	return acl;
}

int chirp_acl_init_root( const char *path )
{
	char aclpath[CHIRP_PATH_MAX];
	char username[USERNAME_MAX];
	FILE *file;

	file = chirp_acl_open(path);
	if(file) {
		chirp_acl_close(file);
		return 1;
	}

	username_get(username);

	sprintf(aclpath,"%s/%s",path,CHIRP_ACL_BASE_NAME);
	file = fopen(aclpath,"w");
	if(file) {
		fprintf(file,"unix:%s %s\n",username,chirp_acl_flags_to_text(CHIRP_ACL_READ|CHIRP_ACL_WRITE|CHIRP_ACL_LIST|CHIRP_ACL_ADMIN));
		fclose(file);
		return 1;
	} else {
		return 0;
	}
}

int chirp_acl_init_copy( const char *path )
{
	char oldpath[CHIRP_LINE_MAX];
	char newpath[CHIRP_LINE_MAX];
	char subject[CHIRP_LINE_MAX];
	FILE *oldfile;
	FILE *newfile;
	int result = 0;
	int flags;

	sprintf(oldpath,"%s/..",path);
	sprintf(newpath,"%s/%s",path,CHIRP_ACL_BASE_NAME);

	oldfile = chirp_acl_open(oldpath);
	if(oldfile) {
		newfile = fopen(newpath,"w");
		if(newfile) {
			while(chirp_acl_read(oldfile,subject,&flags)) {
				fprintf(newfile,"%s %s\n",subject,chirp_acl_flags_to_text(flags));
			}
			fclose(newfile);
			result = 1;
		}
		chirp_acl_close(oldfile);
	}

	return result;
}

int chirp_acl_init_reserve( const char *path, const char *subject )
{
	char aclpath[CHIRP_PATH_MAX];
	FILE *file;
	int aclflags = chirp_acl_get( path, subject ), newflags = 0;

	if(aclflags&CHIRP_ACL_RESERVE_READ)
		newflags|=CHIRP_ACL_READ;
	if(aclflags&CHIRP_ACL_RESERVE_WRITE)
		newflags|=CHIRP_ACL_WRITE;
	if(aclflags&CHIRP_ACL_RESERVE_LIST)
		newflags|=CHIRP_ACL_LIST;
	if(aclflags&CHIRP_ACL_RESERVE_DELETE)
		newflags|=CHIRP_ACL_DELETE;
	if(aclflags&CHIRP_ACL_RESERVE_RESERVE)
		newflags|=CHIRP_ACL_RESERVE;
	if(aclflags&CHIRP_ACL_RESERVE_ADMIN)
		newflags|=CHIRP_ACL_ADMIN;
	if(aclflags&CHIRP_ACL_RESERVE_EXECUTE)
		newflags|=CHIRP_ACL_EXECUTE;

	/*
	compatibility note:
	If no sub-rights are associated with the v right,
	then give all of the subrights.
	*/

	if(newflags==0) newflags = CHIRP_ACL_ALL;

	sprintf(aclpath,"%s/%s",path,CHIRP_ACL_BASE_NAME);
	file = fopen(aclpath,"w");
	if(file) {
		fprintf(file,"%s %s\n",subject,
			chirp_acl_flags_to_text(newflags));
		fclose(file);
		return 1;
	} else {
		return 0;
	}
}

/*
  Because each directory now contains an ACL,
  a simple rmdir will not work on a (perceived) empty directory.
  This function checks to see if the directory is empty,
  save for the ACL file, and deletes it if so.
  Otherwise, it determines an errno and returns with failure.
*/

int chirp_acl_rmdir( const char *path )
{
	DIR *dir;
	struct dirent *d;

	dir = opendir(path);
	if(dir) {
		while((d=readdir(dir))) {
			if(!strcmp(d->d_name,".")) continue;
			if(!strcmp(d->d_name,"..")) continue;
			if(!strcmp(d->d_name,CHIRP_ACL_BASE_NAME)) continue;
			closedir(dir);
			errno = ENOTEMPTY;
			return -1;
		}
		closedir(dir);
		delete_dir(path);
		return 0;
	} else {
		errno = ENOENT;
		return -1;
	}
}
