/*
Copyright (C) 2003-2004 Douglas Thain and the University of Wisconsin
Copyright (C) 2005- The University of Notre Dame
This software is distributed under a BSD-style license.
See the file COPYING for details.

This module written by James Fitzgerald, B.S. 2006.
*/

#ifdef HAS_FUSE

#define _FILE_OFFSET_BITS 64
#define FUSE_USE_VERSION 22

#include <config.h>
#include <fuse.h>

#include "auth_all.h"
#include "chirp_global.h"
#include "chirp_protocol.h"
#include "debug.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <errno.h>
#include <sys/statfs.h>

struct chirp_fuse_dirlist {
	char path[CHIRP_PATH_MAX];
	struct chirp_fuse_dirlist* next;
};

static int chirp_fuse_stoptime = 60;
static struct chirp_fuse_dirlist *head=0;

static void parsepath( const char *path, char *newpath, char *host )
{
	int slashcnt = 0, n=0, pathn=0, serven=0;
	if(!strcmp(path, "/")) // if we're in the root directory
	{
		strcpy(host, "/");
		strcpy(newpath, "/");
	}
	else
	{
		while(path[n] != '\0')
		{
			if(path[n] == '/')
			{
				slashcnt++;
			}

			if(slashcnt == 1 && path[n] != '/') // part of server
			{
				host[serven] = path[n];
				serven++;
			}
			else if(slashcnt > 1)
			{
				newpath[pathn] = path[n];
				pathn++;
			}

			n++;
		}
	}
	newpath[pathn]='\0';
	host[serven] = '\0';
	
	if(slashcnt == 0) // if there are no slashes in the path...i.e. we are lsing the root dir
	{
		strcpy(host, path);
		strcpy(newpath, "/");
	}

}

static int chirp_fuse_getattr(const char *path, struct stat *stbuf)
{
	int result;
	struct chirp_stat buf;
	char newpath[CHIRP_PATH_MAX];
	char host[CHIRP_PATH_MAX];
	parsepath(path,newpath,host); 

	result = chirp_global_stat(host, newpath, &buf, time(0)+chirp_fuse_stoptime );

	stbuf->st_dev = buf.st_dev;
	stbuf->st_ino = buf.st_ino;
	stbuf->st_mode = buf.st_mode;
	stbuf->st_uid = buf.st_uid;
	stbuf->st_gid = buf.st_gid;
	stbuf->st_rdev = buf.st_rdev;
	stbuf->st_size = buf.st_size;
	stbuf->st_blksize = buf.st_blksize;
	stbuf->st_blocks = buf.st_blocks;
	stbuf->st_atime = buf.st_atime;
	stbuf->st_mtime = buf.st_mtime;
	stbuf->st_ctime = buf.st_ctime;

	if(result == -1) return -errno;

	return 0;
}

static int chirp_fuse_readlink(const char *path, char *buf, size_t size)
{
	int result;
	char newpath[CHIRP_PATH_MAX];
	char host[CHIRP_PATH_MAX];
	parsepath(path,newpath,host);
	result = chirp_global_readlink(host, newpath, buf, size, time(0)+chirp_fuse_stoptime );
	
	if(result == -1)
		return -errno;

	buf[result] = '\0';
	return 0;
}

static void chirp_fuse_readdir_callback( const char *path, void *arg )
{
	struct chirp_fuse_dirlist* dirptr = malloc(sizeof(*dirptr));
	strcpy(dirptr->path,path);
	dirptr->next = head;
	head = dirptr;
}

static int chirp_fuse_readdir( const char *path, void *buf, fuse_fill_dir_t filler, off_t offset, struct fuse_file_info *fi )
{
	int result;
 	char newpath[CHIRP_PATH_MAX];
	char host[CHIRP_PATH_MAX];
	struct stat st;
	struct chirp_fuse_dirlist* dirptr = head;
	struct chirp_fuse_dirlist* tmpdir = NULL;

	parsepath(path,newpath,host);

	result = chirp_global_getdir(host, newpath, chirp_fuse_readdir_callback, NULL, time(0)+chirp_fuse_stoptime );

	dirptr = head;
	while(dirptr != NULL)
	{
		char fullpath[CHIRP_PATH_MAX];
		sprintf(fullpath,"%s/%s",path,dirptr->path);
		chirp_fuse_getattr(fullpath,&st);
		filler(buf, dirptr->path,&st, 0);
		tmpdir = dirptr;
		dirptr = dirptr->next;
		free(tmpdir);
	}
	head = NULL;

	if(result == -1)
		return -errno;

	return 0;
}

static int chirp_fuse_mkdir(const char *path, mode_t mode)
{
	int result;
	
	char newpath[CHIRP_PATH_MAX];
	char host[CHIRP_PATH_MAX];
	parsepath(path,newpath,host);
	result = chirp_global_mkdir(host, newpath, mode, time(0) + chirp_fuse_stoptime);
	if(result == -1)
		return -errno;

	return 0;
}

static int chirp_fuse_unlink(const char *path)
{
	int result;
	char newpath[CHIRP_PATH_MAX];
	char host[CHIRP_PATH_MAX];
	parsepath(path,newpath,host);
	result = chirp_global_unlink(host, newpath, time(0) + chirp_fuse_stoptime );
	if(result == -1)
		return -errno;

	return 0;
}

static int chirp_fuse_rmdir(const char *path)
{
	int result;
	char newpath[CHIRP_PATH_MAX];
	char host[CHIRP_PATH_MAX];
	parsepath(path,newpath,host); 
	result = chirp_global_rmdir(host, newpath, time(0) + chirp_fuse_stoptime );
	if(result == -1)
		return -errno;

	return 0;
}

static int chirp_fuse_symlink(const char *from, const char *to)
{
	int result;
	char frompath[CHIRP_PATH_MAX], topath[CHIRP_PATH_MAX];
	char host[CHIRP_PATH_MAX];
	parsepath(from,frompath,host); 
	parsepath(to,topath,host);
	result = chirp_global_symlink( host, frompath, topath, time(0)+ chirp_fuse_stoptime );

	if(result == -1)
		return -errno;

	return 0;
}

static int chirp_fuse_rename(const char *from, const char *to)
{
	int result;
	char frompath[CHIRP_PATH_MAX], topath[CHIRP_PATH_MAX];
	char host[CHIRP_PATH_MAX];
	parsepath(from,frompath,host); 
	parsepath(to,topath,host);
	result = chirp_global_rename( host, frompath, topath, time(0)+ chirp_fuse_stoptime );

	if(result == -1)
		return -errno;

	return 0;
}

static int chirp_fuse_link(const char *from, const char *to)
{
	int result;
	char frompath[CHIRP_PATH_MAX], topath[CHIRP_PATH_MAX];
	char host[CHIRP_PATH_MAX];
	parsepath(from,frompath,host); 
	parsepath(to,topath,host);
	result = chirp_global_link( host, frompath, topath, time(0)+ chirp_fuse_stoptime );
	if(result == -1)
		return -errno;

	return 0;
}

static int chirp_fuse_chmod(const char *path, mode_t mode)
{
	int result;
	char newpath[CHIRP_PATH_MAX];
	char host[CHIRP_PATH_MAX];
	parsepath(path,newpath,host);
	result = chirp_global_chmod( host, newpath, mode, time(0)+chirp_fuse_stoptime );
	if(result == -1)
		return -errno;

	return 0;
}

static int chirp_fuse_chown(const char *path, uid_t uid, gid_t gid)
{
	int result;
	char newpath[CHIRP_PATH_MAX];
	char host[CHIRP_PATH_MAX];
	parsepath(path,newpath,host);
	result = chirp_global_chown( host, newpath, uid, gid, time(0)+chirp_fuse_stoptime );

	if(result == -1)
		return -errno;

	return 0;
}

static int chirp_fuse_truncate(const char *path, off_t size)
{
	int result;
	char newpath[CHIRP_PATH_MAX];
	char host[CHIRP_PATH_MAX];
	parsepath(path,newpath,host);
	result = chirp_global_truncate( host, newpath, size, time(0)+chirp_fuse_stoptime );

	if(result == -1) return -errno;

	return 0;
}

static int chirp_fuse_utime(const char *path, struct utimbuf *buf)
{
	int result;
	char newpath[CHIRP_PATH_MAX];
	char host[CHIRP_PATH_MAX];
	parsepath(path,newpath,host);
	time_t actime, modtime;

	result = chirp_global_utime( host, newpath, actime, modtime, time(0)+chirp_fuse_stoptime );

	buf->actime = actime;
	buf->modtime = modtime;
	
	if(result == -1) return -errno;

	return 0;
}


static int chirp_fuse_open( const char *path, struct fuse_file_info *fi )
{
	struct chirp_file *file;
	int mode = 0;
	char newpath[CHIRP_PATH_MAX];
	char host[CHIRP_PATH_MAX];
	parsepath(path,newpath,host); 
	file =  chirp_global_open( host, newpath, fi->flags, mode, time(0)+chirp_fuse_stoptime);
	if(file == NULL) return -errno;

	chirp_global_close(file, time(0) + chirp_fuse_stoptime);

	return 0;

}

static int chirp_fuse_read(const char *path, char *buf, size_t size, off_t offset, struct fuse_file_info *fi)
{
	struct chirp_file * file;
	int result;
	int mode = 0;
	char newpath[CHIRP_PATH_MAX];
	char host[CHIRP_PATH_MAX];
	parsepath(path,newpath,host); 
	file = chirp_global_open( host, newpath, O_RDONLY, mode, time(0)+chirp_fuse_stoptime);
	if(file == NULL) return -errno;

	result = chirp_global_pread( file, buf, size, offset, time(0)+chirp_fuse_stoptime );

	chirp_global_close(file, time(0) + chirp_fuse_stoptime);
	return result;

}

static int chirp_fuse_write(const char *path, const char *buf, size_t size, off_t offset, struct fuse_file_info *fi)
{
	struct chirp_file * file;
	int result;
	int mode = 0;
	char newpath[CHIRP_PATH_MAX];
	char host[CHIRP_PATH_MAX];
	parsepath(path,newpath,host);
	file = chirp_global_open( host, newpath, O_WRONLY, mode, time(0)+chirp_fuse_stoptime);

	if(file == NULL) return -errno;

	result = chirp_global_pwrite( file, buf, size, offset, time(0)+chirp_fuse_stoptime );
	if(result == -1) result = -errno;

	chirp_global_close(file, time(0) + chirp_fuse_stoptime);
	return result;
}

static int chirp_fuse_mknod(const char *path, mode_t mode, dev_t rdev)
{
	struct chirp_file *result;
	char newpath[CHIRP_PATH_MAX];
	char host[CHIRP_PATH_MAX];
	parsepath(path,newpath,host);
	result =  chirp_global_open( host, newpath, O_CREAT, mode, time(0)+chirp_fuse_stoptime);

	if(result == NULL)
		return -errno;

	return 0;
}


static int chirp_fuse_statfs(const char *path, struct statfs *stbuf)
{
	int result;
	char newpath[CHIRP_PATH_MAX];
	char host[CHIRP_PATH_MAX];
	parsepath(path,newpath,host);
	struct chirp_statfs buf;

	result = chirp_global_statfs( host, newpath, &buf, time(0)+chirp_fuse_stoptime );

	memset(stbuf,0,sizeof(*stbuf));

	stbuf->f_type = buf.f_type;
	stbuf->f_bsize = buf.f_bsize;
	stbuf->f_blocks = buf.f_blocks;
	stbuf->f_bfree = buf.f_bfree;
	stbuf->f_bavail = buf.f_bavail;
	stbuf->f_files = buf.f_files;
	stbuf->f_ffree = buf.f_ffree;

	if(result == -1)
		return -errno;

	return 0;
}

static struct fuse_operations chirp_fuse_operations = {
	.getattr	= chirp_fuse_getattr,
	.readlink	= chirp_fuse_readlink,
	.readdir	= chirp_fuse_readdir,
	.mknod		= chirp_fuse_mknod,
	.mkdir		= chirp_fuse_mkdir,
	.symlink	= chirp_fuse_symlink,
	.unlink		= chirp_fuse_unlink,
	.rmdir		= chirp_fuse_rmdir,
	.rename		= chirp_fuse_rename,
	.link		= chirp_fuse_link,
	.chmod		= chirp_fuse_chmod,
	.chown		= chirp_fuse_chown,
	.truncate	= chirp_fuse_truncate,
	.utime		= chirp_fuse_utime,
	.open		= chirp_fuse_open,
	.read		= chirp_fuse_read,
	.write		= chirp_fuse_write,
	.statfs		= chirp_fuse_statfs,
};

int main(int argc, char *argv[])
{
	char *flags = getenv("CHIRP_FUSE_DEBUG");
	debug_config(argv[0]);
	debug_config_file("/dev/stdout");
	if(flags) debug_flags_set(flags);
	auth_register_all();
	return fuse_main(argc, argv, &chirp_fuse_operations);
}

#else

int main( int argc, char *argv[] )
{
	printf("%s: sorry, fuse support was not built in\n",argv[0]);
	return 1;
}

#endif
