
#include "chirp_multi.h"
#include "chirp_reli.h"
#include "chirp_protocol.h"

#include "debug.h"
#include "stringtools.h"
#include "create_dir.h"
#include "hash_table.h"
#include "xmalloc.h"

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <sys/stat.h>
#ifdef HAS_SYS_STATFS_H
#include <sys/statfs.h>
#endif
#include <sys/mount.h>
#include <sys/param.h>
#include <dirent.h>
#include <utime.h>
#include <pwd.h>

struct chirp_file {
	char host[CHIRP_LINE_MAX];
	char path[CHIRP_LINE_MAX];
	struct chirp_stat info;
	INT64_T fd;
	INT64_T flags;
	INT64_T mode;
	INT64_T serial;
    INT64_T stale;
};

struct chirp_volume *current_volume=0;

struct file_info {
	char lpath[CHIRP_PATH_MAX];
	char rpath[CHIRP_PATH_MAX];
	char rhost[CHIRP_PATH_MAX];
};

struct chirp_server {
	char name[CHIRP_PATH_MAX];
	int priority;
	int prepared;
        int known_bad;
	struct chirp_server *next;
};

struct chirp_volume {
	char name[CHIRP_PATH_MAX];
	char host[CHIRP_PATH_MAX];
	char root[CHIRP_PATH_MAX];
	char key[CHIRP_PATH_MAX];
	struct chirp_server *list;
	struct chirp_server *choice;
};

struct chirp_volume * chirp_volume_open( const char *volume, time_t stoptime )
{
	struct chirp_volume *v;
	char filename[CHIRP_PATH_MAX];
	char *buffer;
	char *name;
	char *c;
	int result;
	int offset;
	int num_servers = 0;
	/*
	  The volume name must have at least one
	  at-sign in order to indicate the volname.
	*/

	c = strchr(volume,'@');
	if(!c) {
		errno = ENOENT;
		return 0;
	}

	v = xxmalloc(sizeof(*v));
	strcpy(v->name,volume);
	v->list = 0;
	v->choice = 0;

	/*
	  The stuff preceding the first at-sign is
	  the name of the host containing the directory.
	*/

	strcpy(v->host,v->name);
	c = strchr(v->host,'@');

	/* The rest is the logical name of the volume. */
	strcpy(v->root,c);
	*c = 0;

	/* Convert any remaning at-signs to slashes */
	for(c=v->root;*c;c++) {
		if(*c=='@') *c = '/';
	}

	/* Fetch the filesystem key */
	sprintf(filename,"%s/key",v->root);
	result = chirp_reli_getfile_buffer(v->host,filename,&buffer,stoptime);
	if(result<0) {
		debug(D_CHIRP,"couldn't open %s: %s\n",filename,strerror(errno));
		free(v);
		errno = ENOENT;
		return 0;
	}
	strcpy(v->key,buffer);
	string_chomp(v->key);
	free(buffer);

	/* Fetch the list of hosts */
	sprintf(filename,"%s/hosts",v->root);
	result = chirp_reli_getfile_buffer(v->host,filename,&buffer,stoptime);
	if(result<0) {
		debug(D_CHIRP,"couldn't open %s: %s\n",filename,strerror(errno));
		free(v);
		errno = ENOENT;
		return 0;
	}

	/* The list is delimited by whitespace */
	name = strtok(buffer," \t\n");
	while(name) {
		num_servers++;
		struct chirp_server *s = xxmalloc(sizeof(*s));
		strcpy(s->name,name);
		debug(D_MULTI,"adding host %s", name);
		s->priority = 0;
		s->prepared = 0;
		s->known_bad = 0;
		s->next = v->list;
		v->list = s;
		name = strtok(0," \t\n");
	}

	/* Set the initial choice to be a random server in the list */
	offset = (int) (num_servers*(rand()/(RAND_MAX + 1.0)));  //pseudo-random number between 0 and num_servers-1, inclusive.
	v->choice = v->list;
	for(num_servers = 0; num_servers < offset; num_servers++)
	{
		if(v->choice) v->choice = v->choice->next;
		if(!v->choice) {
			v->choice = v->list;
			if(!v->choice)
			{
				debug(D_CHIRP,"volume %s has no hosts available!\n",v->name);
			}
		}		
	}
	free(buffer);

	return v;
}

void chirp_volume_close( struct chirp_volume *v )
{
	struct chirp_server *s;
	
	while(v->list) {
		s = v->list->next;
		free(v->list);
		v->list = s;
	}
	
	free(v);		
}

struct chirp_server * volume_choose_memo(struct chirp_volume *v)
{
	/* MEMOIZATION CODE!*/
	struct chirp_server * curr;
	if(v->choice) v->choice = v->choice->next;
	if(!v->choice) v->choice = v->list;
	
	
	if(v->choice) { // if either the next or first node isn't null.
		curr = v->choice; // set curr to choice
		if(curr->known_bad == 0) // if choice is okay, use it
		{
			debug(D_MULTI, "first chosen server %s is not known to be bad.", curr->name);
			return v->choice;
		}
		else {
			debug(D_MULTI, "first chosen server %s is known to be bad, trying others:", curr->name);
			do {
				curr = curr->next; // otherwise get next node
				if(!curr) curr = v->list; // wrap around if necessary
				debug(D_MULTI, "testing server %s to see if it is known to be bad", curr->name); 
			} while(curr != v->choice && curr->known_bad == 1); // loop until good node, or back to choice.
			if(curr->known_bad == 0) { // if we found a possibly good node, choose it.
				debug(D_MULTI, "Finally chose %s", curr->name); 
				v->choice = curr;
				return v->choice;
			}
			else {
				// got back to choice without finding a good node.
				// reset entire list to try again.
				// and return what would have been the next node in the first place.
				debug(D_MULTI, "all servers have been eliminated."); 
				curr = v->list;
				while(curr)
				{
					curr->known_bad = 0;
					curr = curr->next;
				}
				debug(D_MULTI, "Starting anew choosing %s", curr->name); 
				return v->choice;
			}
		} // end "else -> if choice wasn't okay"	
	} else {
		debug(D_CHIRP,"volume %s has no hosts available!\n",v->name);
		return 0;
	}	
}

struct chirp_server * chirp_volume_choose( struct chirp_volume *v  )
{

	/*ENV CODE*/
	struct chirp_server* orig_choice;
	struct chirp_server* env_new;
	//struct chirp_stat *buf = xxmalloc(sizeof(*buf));
	char *require_host = getenv("CHIRP_MULTI_REQUIRE");
	char *request_host = getenv("CHIRP_MULTI_REQUEST");
	char *envhost = NULL;
	int require = 0;
	int request = 0;
	if(request_host != NULL)
	{
		envhost = request_host;
		request = 1;
	}
	if(require_host != NULL) {
		envhost = require_host;
		require = 1;
	}
		
	if(require || request)
	{
		debug(D_MULTI, "Will attempt to place file on host %s\n", envhost); 
		if(v->choice) v->choice = v->choice->next;
		if(!v->choice) {
			v->choice = v->list;
			if(!v->choice)
			{
				debug(D_CHIRP,"volume %s has no hosts available!\n",v->name);
				return 0;
			}
		}
		
		orig_choice = v->choice;
		v->choice = v->choice->next;
		if(!v->choice) v->choice = v->list;
		// check if this is it.
		if(!strcmp(orig_choice->name,envhost)) {
			debug(D_MULTI, "%s was found; matching %s\n", orig_choice->name, envhost);
			if(orig_choice->known_bad == 0 || require)
				return orig_choice;
			else // was only a request, and orig_choice is known bad, so just pick a random host
				return volume_choose_memo(v);
		}
		// otherwise, loop until found, or back to starting point.
		while(v->choice != orig_choice) {
			if(!strcmp(v->choice->name,envhost)) {
				debug(D_MULTI, "%s was found; matching %s \n", v->choice->name, envhost);
				if(v->choice->known_bad == 0 || require)
					return v->choice;
				else {// was only a request, and choice is known bad, so just pick a random host

                                        // THIS DOES NOT CONFORM TO THE COMMENT; IT JUST PICKS THE HOST IMMEDIATELY AFTER THE REQUESTED ONE!!!!!
					// Implications are possible unbalanced writes: when the requested server has failed, all writes go to the
					// next server, rather than being spread acrsoss the array.
					debug(D_MULTI, "%s is known to be bad. Choosing a different host.", v->choice->name);
					return volume_choose_memo(v);
				}
			} else
			{
				debug(D_MULTI, "%s does not match %s. Trying next ...\n", v->choice->name, envhost);
				v->choice = v->choice->next;
				if(!v->choice) v->choice = v->list;
			}
		}
		// else, we didn't find the host, so we try the host as a new possibility.
		debug(D_MULTI, "Could not find %s.  Creating it.", envhost);
		env_new = xxmalloc(sizeof(*env_new));
		strcpy(env_new->name,envhost);
		if(require)
			env_new->priority = 2;
		else
			env_new->priority = 1;
		env_new->prepared = 0;
		env_new->known_bad = 0;
		// This will prevent Parrot from infinite looping on an unknown host,
		// but relies on a magic #, and the assumption that if you can't stat it now, it doesn't exist.
		/*
		  if(chirp_reli_stat( envhost, "/", buf, time(0)+5) < 0)
		  {
		  debug(D_MULTI,"The environment variable supplied a chirp server I can't reach!");
		  free(buf);
		  free(env_new);
		  return 0;
		  }
		  else{
		  free (buf);
		  }
		*/
		env_new->next = v->list;
		v->list = env_new;
		v->choice = env_new;
		return v->choice;
	}
	else { // no environment variable set ... do standard memoization.
		return volume_choose_memo(v);
	}
}

static int chirp_multi_init( const char *volume, time_t stoptime)
{
	char *c;

	c = strrchr(volume,':');
	if(c) *c = 0;

	if(current_volume && strcmp(current_volume->name,volume)) {
		chirp_volume_close(current_volume);
		current_volume = 0;
	}

	if(!current_volume) {
		current_volume = chirp_volume_open(volume,stoptime);
		if(!current_volume) return 0;
	}

	return 1;
}

static int chirp_multi_lpath( const char *volume, const char *path, char *lpath, time_t stoptime )
{
	/*
	  Hack: df tries to search above /multi/volume/..
	  and then gets confused.  Stop it at the root.
	*/
	if(!strncmp(path,"/..",3)) path="/";

	if(!chirp_multi_init(volume,stoptime)) return 0;
	sprintf(lpath,"%s/root/%s",current_volume->root,path);
	return 1;
}

static int chirp_multi_lookup( const char *volume, const char *path, struct file_info *info, time_t stoptime )
{
	int result, fields;
	char *buffer;

	if(!chirp_multi_lpath(volume,path,info->lpath,stoptime)) return 0;

	result = chirp_reli_getfile_buffer(current_volume->host,info->lpath,&buffer,stoptime);
	if(result<=0) return 0;

	fields = sscanf(buffer,"%s %s",info->rhost,info->rpath);

	free(buffer);

	if(fields==2) {
		return 1;
	} else {
		errno = EIO;
		return 0;
	}
}
		
static int chirp_multi_update( const char *volume, const char *path, struct file_info *info, time_t stoptime )
{
	char buffer[CHIRP_PATH_MAX*2+2];
	int result;
	if(!chirp_multi_lpath(volume,path,info->lpath,stoptime)) return 0;

	sprintf(buffer,"%s\n%s\n",info->rhost,info->rpath);
	result = chirp_reli_putfile_buffer(current_volume->host,info->lpath,buffer,0700,strlen(buffer),stoptime);

	if(result==strlen(buffer)) {
		return 1;
	} else {
		return 0;
	}
}
		
static void chirp_multi_generate( char *s )
{
	int i;
	for(i=0;i<16;i++) {
		s[i] = 'a'+(rand()>>8)%26;
	}
	s[i] = 0;
}

static int chirp_multi_create( const char *volume, const char *path, struct file_info *info, time_t stoptime )
{

        struct chirp_server *choice;
	int result;
	char tokpath[CHIRP_PATH_MAX];
	char tokkey[CHIRP_PATH_MAX];
	char* tokdir;
	
	debug(D_MULTI,"Creating (v)%s(p)%s (lp)%s", volume, path, info->lpath);
	
	if(!chirp_multi_lpath(volume,path,info->lpath,stoptime)) return 0;
	
	while(1) {
		choice = chirp_volume_choose(current_volume);
		if(!choice) {
			errno = ENOSPC;
			return 0;
		}
		
		if(!choice->prepared) {
			/*Chirp doesn't like creating multiple levels of directories in one mkdir
			  so chop up the key path and create one at a time*/
			tokpath[0] = '\0';
			strcpy(tokkey, current_volume->key);
			debug(D_MULTI,"tokenizing: %s",tokkey);
			tokdir = strtok(tokkey, "/");
			while(tokdir)
			{
				strcat(tokpath, "/");
				strcat(tokpath, tokdir);
				debug(D_MULTI,"making path: %s",tokpath);
				result = 0;
				result = chirp_reli_mkdir(choice->name,tokpath,0777,time(0)+5);
				if(result!=0 && errno!=EEXIST) { // failed for reason other than already existing
					break; // stop tokenizing
				}
				tokdir = strtok(0,"/");
			}
			if(result==0 || (result==-1 && errno==EEXIST)) { // if we didn't fail on the last try, we didn't break out early
				choice->prepared = 1;
				break;
			} else { // we failed somewhere along the way.
				if(choice->priority != 2) {
					debug(D_NOTICE,"server %s is not responding, looking for another..",choice->name);
					debug(D_MULTI, "set %s to known_bad", choice->name);
					choice->known_bad = 1;
					continue;
				}
				else {
					debug(D_NOTICE,"server %s is not responding, returning null ...",choice->name);
					return 0;
					
				}
			}
			tokdir = strtok(0,"/");
		}
		else { // already prepared.
			break;
		}
	}

	strcpy(info->rhost,choice->name);
	strcpy(info->rpath,current_volume->key);
	strcat(info->rpath,"/");
	chirp_multi_generate(&info->rpath[strlen(info->rpath)]);
	return chirp_multi_update(volume,path,info,stoptime);
}

struct chirp_file * chirp_multi_open( const char *volume, const char *path, INT64_T flags, INT64_T mode, time_t stoptime )
{
	struct file_info info;
	struct chirp_file* retfile;
	//INT64_T retval;

	if(!chirp_multi_lookup(volume,path,&info,stoptime)) {
		if(errno==ENOENT && flags&O_CREAT) {
			if(!chirp_multi_create(volume,path,&info,stoptime)) return 0;
		} else {
			return 0;
		}
	}
        //return chirp_reli_open(info.rhost,info.rpath,flags,mode,stoptime);
	
	retfile = chirp_reli_open(info.rhost,info.rpath,flags,mode,stoptime);
	while(retfile == NULL) {
	    debug(D_MULTI," M: errno=%i (%s)",errno, strerror(errno));
	    if(errno == EACCES)
	    {
		debug(D_MULTI,"Marking %s as known_bad", current_volume->choice->name);
		current_volume->choice->known_bad = 1;
                // delete the stub file
		debug(D_MULTI,"Trying to delete stub file");
		errno = 0;
		if(chirp_multi_unlink(volume, path, stoptime) == -1)
		{
		    debug(D_MULTI,"Failed to delete stub file! Crap!");
		    debug(D_MULTI,"Delete Stub File: errno=%i (%s)",errno, strerror(errno));
		    return retfile;
		}
                // redo stuff above
		debug(D_MULTI,"Trying again");
		if(!chirp_multi_lookup(volume,path,&info,stoptime)) {
		    if(errno==ENOENT && flags&O_CREAT) {
			if(!chirp_multi_create(volume,path,&info,stoptime)) return 0;
		    } else {
			return 0;
		    }
		}
		retfile = chirp_reli_open(info.rhost,info.rpath,flags,mode,stoptime);
	    }
	    else
	    {
		debug(D_MULTI," returning NULL with some other error");
		return retfile;
	    }
	}
	
	return retfile;
	
}

INT64_T chirp_multi_close( struct chirp_file *file, time_t stoptime )
{
	return chirp_reli_close(file,stoptime);
}

INT64_T chirp_multi_pread( struct chirp_file *file, void *buffer, INT64_T length, INT64_T offset, time_t stoptime )
{
	return chirp_reli_pread(file,buffer,length,offset,stoptime);
}

INT64_T chirp_multi_pwrite( struct chirp_file *file, const void *buffer, INT64_T length, INT64_T offset, time_t stoptime )
{
    debug(D_MULTI,"pwrite %lld bytes at offset %lld", length, offset);
    return chirp_reli_pwrite(file,buffer,length,offset,stoptime);
}

INT64_T chirp_multi_fstat( struct chirp_file *file, struct chirp_stat *buf, time_t stoptime )
{
	return chirp_reli_fstat(file,buf,stoptime);
}

INT64_T chirp_multi_fstatfs( struct chirp_file *file, struct chirp_statfs *buf, time_t stoptime )
{
	return chirp_reli_fstatfs(file,buf,stoptime);
}

INT64_T chirp_multi_fchown( struct chirp_file *file, INT64_T uid, INT64_T gid, time_t stoptime )
{
	return chirp_reli_fchown(file,uid,gid,stoptime);
}

INT64_T chirp_multi_fchmod( struct chirp_file *file, INT64_T mode, time_t stoptime )
{
	return chirp_reli_fchmod(file,mode,stoptime);
}

INT64_T chirp_multi_ftruncate( struct chirp_file *file, INT64_T length, time_t stoptime )
{
	return chirp_reli_ftruncate(file,length,stoptime);
}

INT64_T chirp_multi_getfile( const char *volume, const char *path, FILE *stream, time_t stoptime )
{
	struct file_info info;
	if(!chirp_multi_lookup(volume,path,&info,stoptime)) return -1;
	return chirp_reli_getfile(info.rhost,info.rpath,stream,stoptime);
}

INT64_T chirp_multi_putfile( const char *volume, const char *path, FILE *stream, INT64_T mode, INT64_T length, time_t stoptime )
{
	struct file_info info;
	debug(D_MULTI,"putfile %s%s put %lld bytes", volume, path, length);
	if(!chirp_multi_lookup(volume,path,&info,stoptime)) {
		if(!chirp_multi_create(volume,path,&info,stoptime)) return -1;
	}
	return chirp_reli_putfile(info.rhost,info.rpath,stream,mode,length,stoptime);
}

INT64_T chirp_multi_getdir( const char *volume, const char *path, chirp_dir_t callback, void *arg, time_t stoptime )
{
	char lpath[CHIRP_PATH_MAX];
	if(!chirp_multi_lpath(volume,path,lpath,stoptime)) return -1;
	return chirp_reli_getdir(current_volume->host,lpath,callback,arg,stoptime);
}

INT64_T chirp_multi_unlink( const char *volume, const char *path, time_t stoptime )
{
	struct file_info info;
	int result;

	if(chirp_multi_lookup(volume,path,&info,stoptime)) {
		result = chirp_reli_unlink(info.rhost,info.rpath,stoptime);
		if(result!=0 && (errno!=ENOENT && errno!=EACCES))
		{
		    debug(D_MULTI,"Unlink1: errno=%i (%s)",errno, strerror(errno));
		    return -1;
		}
	}

	result = chirp_reli_unlink(current_volume->host,info.lpath,stoptime);
	if(result!=0)
	{
	    debug(D_MULTI,"Unlink2: errno=%i (%s)",errno, strerror(errno));
	    return -1;
	}
	return result;
}

INT64_T chirp_multi_rename( const char *volume, const char *path, const char *newpath, time_t stoptime )
{
	char lpath[CHIRP_PATH_MAX];
	char newlpath[CHIRP_PATH_MAX];
	if(!chirp_multi_lpath(volume,path,lpath,stoptime)) return -1;
	if(!chirp_multi_lpath(volume,newpath,newlpath,stoptime)) return -1;
	return chirp_reli_rename(current_volume->host,lpath,newlpath,stoptime);
}

INT64_T chirp_multi_link( const char *volume, const char *path, const char *newpath, time_t stoptime )
{
	errno = ENOSYS;
	return -1;
}

INT64_T chirp_multi_symlink( const char *volume, const char *path, const char *newpath, time_t stoptime )
{
	char lpath[CHIRP_PATH_MAX];
	char newlpath[CHIRP_PATH_MAX];
	if(!chirp_multi_lpath(volume,path,lpath,stoptime)) return -1;
	if(!chirp_multi_lpath(volume,newpath,newlpath,stoptime)) return -1;
	return chirp_reli_symlink(current_volume->host,lpath,newlpath,stoptime);
}

INT64_T chirp_multi_readlink( const char *volume, const char *path, char *buf, INT64_T length, time_t stoptime )
{
	char lpath[CHIRP_PATH_MAX];
	if(chirp_multi_lpath(volume,path,lpath,stoptime)) {
		return chirp_reli_readlink(current_volume->host,lpath,buf,length,stoptime);
	} else {
		return -1;
	}
}

INT64_T chirp_multi_mkdir( const char *volume, char const *path, INT64_T mode, time_t stoptime )
{
	char lpath[CHIRP_PATH_MAX];
	if(chirp_multi_lpath(volume,path,lpath,stoptime)) {
		return chirp_reli_mkdir(current_volume->host,lpath,mode,stoptime);
	} else {
		return -1;
	}
}

INT64_T chirp_multi_rmdir( const char *volume, char const *path, time_t stoptime )
{
	char lpath[CHIRP_PATH_MAX];
	if(chirp_multi_lpath(volume,path,lpath,stoptime)) {
		return chirp_reli_rmdir(current_volume->host,lpath,stoptime);
	} else {
		return -1;
	}
}

INT64_T chirp_multi_stat( const char *volume, const char *path, struct chirp_stat *buf, time_t stoptime )
{
	struct file_info info;

	if(chirp_multi_lookup(volume,path,&info,stoptime)) {
		return chirp_reli_stat(info.rhost,info.rpath,buf,stoptime);
	} else if(errno==EISDIR) {
		return chirp_reli_stat(current_volume->host,info.lpath,buf,stoptime);
	} else {
		return -1;
	}
}

INT64_T chirp_multi_lstat( const char *volume, const char *path, struct chirp_stat *buf, time_t stoptime )
{
	struct file_info info;

	if(chirp_multi_lookup(volume,path,&info,stoptime)) {
		return chirp_reli_lstat(info.rhost,info.rpath,buf,stoptime);
	} else if(errno==EISDIR) {
		return chirp_reli_lstat(current_volume->host,info.lpath,buf,stoptime);
	} else {
		return -1;
	}
}

INT64_T chirp_multi_statfs( const char *volume, const char *path, struct chirp_statfs *buf, time_t stoptime )
{
	struct chirp_server *s;
	INT64_T files_total = 0;
	INT64_T bytes_avail = 0;
	INT64_T bytes_free = 0;
	INT64_T bytes_total = 0;
	INT64_T files_free = 0;
	INT64_T block_size = 4096;

	if(!chirp_multi_init(volume,stoptime)) return 0;

	for(s=current_volume->list;s;s=s->next) {
		struct chirp_statfs tmp;
		int result;

		result = chirp_reli_statfs(s->name,"/",&tmp,stoptime);
		if(result<0) return result;

		bytes_total += tmp.f_blocks * (INT64_T)tmp.f_bsize;
		bytes_avail += tmp.f_bavail * (INT64_T)tmp.f_bsize;
		bytes_free  += tmp.f_bfree * (INT64_T)tmp.f_bsize;
		files_total += tmp.f_files;
		files_free  += tmp.f_ffree;
	}

	memset(buf,0,sizeof(*buf));

	buf->f_bsize = block_size;
	buf->f_blocks = bytes_total/block_size;
	buf->f_bavail = bytes_avail/block_size;
	buf->f_bfree = bytes_free/block_size;
	buf->f_files = files_total;
	buf->f_ffree = files_total;

	return 0;
}

INT64_T chirp_multi_access( const char *volume, const char *path, INT64_T mode, time_t stoptime )
{
	struct file_info info;
	if(chirp_multi_lookup(volume,path,&info,stoptime)) {
		return chirp_reli_access(info.rhost,info.rpath,mode,stoptime);
	} else if(errno==EISDIR) {
		return chirp_reli_access(current_volume->host,info.lpath,mode,stoptime);
	} else {
		return -1;
	}
}

INT64_T chirp_multi_chmod( const char *volume, const char *path, INT64_T mode, time_t stoptime )
{
	struct file_info info;
	if(chirp_multi_lookup(volume,path,&info,stoptime)) {
		return chirp_reli_chmod(info.rhost,info.rpath,mode,stoptime);
	} else if(errno==EISDIR) {
		return chirp_reli_chmod(current_volume->host,info.lpath,mode,stoptime);
	} else {
		return -1;
	}
}

INT64_T chirp_multi_chown( const char *volume, const char *path, INT64_T uid, INT64_T gid, time_t stoptime )
{
	struct file_info info;
	if(chirp_multi_lookup(volume,path,&info,stoptime)) {
		return chirp_reli_chown(info.rhost,info.rpath,uid,gid,stoptime);
	} else if(errno==EISDIR) {
		return chirp_reli_chown(current_volume->host,info.lpath,uid,gid,stoptime);
	} else {
		return -1;
	}
}

INT64_T chirp_multi_lchown( const char *volume, const char *path, INT64_T uid, INT64_T gid, time_t stoptime )
{
	struct file_info info;
	if(chirp_multi_lookup(volume,path,&info,stoptime)) {
		return chirp_reli_lchown(info.rhost,info.rpath,uid,gid,stoptime);
	} else if(errno==EISDIR) {
		return chirp_reli_lchown(current_volume->host,info.lpath,uid,gid,stoptime);
	} else {
		return -1;
	}
}

INT64_T chirp_multi_truncate( const char *volume, const char *path, INT64_T length, time_t stoptime )
{
	struct file_info info;
	if(chirp_multi_lookup(volume,path,&info,stoptime)) {
		return chirp_reli_truncate(info.rhost,info.rpath,length,stoptime);
	} else {
		return -1;
	}
}

INT64_T chirp_multi_utime( const char *volume, const char *path, time_t actime, time_t modtime, time_t stoptime )
{
	struct file_info info;
	if(chirp_multi_lookup(volume,path,&info,stoptime)) {
		return chirp_reli_utime(info.rhost,info.rpath,actime,modtime,stoptime);
	} else if(errno==EISDIR) {
		return chirp_reli_utime(current_volume->host,info.lpath,actime,modtime,stoptime);
	} else {
		return -1;
	}
}
